package org.openscm.kundo.plugins

/*
 * Copyright (C) 2008 The Ultimate People Company Ltd ("UPCO").
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import org.openscm.kundo.plugins.context.BuildContext

/**
 * junit Delegate
 * @author <a href="mailto:steve.king@upco.co.uk">Steve king</a>
 * @version 1.0.0
 * 
 * Description: The junit plugin calls Junit batch tests for each output directory set by the kundo-project-plugin.
 * 
 * <p>The junit delegate consumes data set by the kundo-project-plugin and uses each test output path to run batch
 * tests. Single unit tests can be run on classes defined as a comma separated list.</p>
 */
/* 
 * @ant-target doJunit
 */
class JunitDelegate extends AbstractPluginTargetDelegate {

	/*
	 * An example set of test setting overrides for indexed properties. If no junit indexed properties are set all defaults will be used. Format: [plugin].[key].[index].[fieldname]
	 * @indexed-ant-property test
	 */	
	private Map indexedProps
	
	/*
	 * Additional individual test can be defined using the 'junit.test.classes' variable which expects a comma seperated list of classes to test. 
	 * @ant-property-name junit.test.classes
	 */	
	private String testClasses
	
	// properties set by the kundo-project-plugin.
	private projectProperties
	
	/*
	 * The location to store junit reports.
	 * @ant-property-name junit.default.reportdir
	 */	
	private defaultReportdir
	
	/*
	 * The file include pattern for the junit tests.
	 * @ant-property-name junit.default.test.includepattern
	 */		
	private defaultTestIncludepattern
	
	/*
	 * The file exclude pattern for the junit tests.
	 * @ant-property-name junit.default.test.excludepattern
	 */			
	private defaultTestExcludepattern
	
	/*
	 * Property to specify if junit summary should be printed.
	 * @ant-property-name junit.default.report.printsummary
	 */		
	private defaultReportPrintsummary
	
	/*
	 * The type of reports junit should generate.
	 * @ant-property-name junit.default.report.formatter
	 */		
	private defaultReportFormatter

	/*
	 * Runs the junit test in separate VM.
	 * @ant-property-name junit.default.fork
	 */		
	private String defaultFork
	/*
	 * Controls how many Java Virtual Machines get created if you want to fork some tests. Possible values are "perTest" (the default), "perBatch" and "once".
	 * @ant-property-name junit.default.fork.mode
	 */		
	private String defaultForkMode
	
	/**
	 * This id will be fetched as an Ant RefId and prefixed to the start of the junit execution classpath. This can be used when integrating to test coverage 
	 * frameworks which require the instrumented code to be made available on the classpath
	 * @ant-property-name junit.pre.classpath.id
	 */
	private String preClasspathId

	/**
	 * This id will be fetched as an Ant RefId and appended to the start of the junit execution classpath. 
	 * @ant-property-name junit.post.classpath.id
	 */
	private String postClasspathId	
	
	/**
	 * Constructor sets ant and buildContext instances in super class
	 * @param ant			AntBuilder instance
	 * @param buildContext	BuildContext instance
	 */	
	JunitDelegate( AntBuilder ant, BuildContext buildContext ){
		super( ant, buildContext)
	}
	
	/**
	 * Implementation of abstract method from AbstractPluginTargetDelegate
	 * to hold assertions that all global and local properties are available to the plugin.
	 * This method is automatically called prior to execution.
	 */	
	void doCheck(){

		// assert global variables exists in ant project
		assert buildContext.get( "project.properties" ) != null, "The project properties could not be retrieved. Please ensure that the project metadata has been loaded via the project plugin"
		// Check that the project properties were injected
		assert indexedProps != null
		assert testClasses != null
		assert getSysKeyOptions() != null

		assert defaultFork != null, 		"Error due to NULL value of junit.fork"
		assert defaultForkMode != null, 	"Error due to NULL value of junit.fork.mode"
	}	
	
	/**
	 * Setter method for member variable indexedProps
	 * @param indexedProps	Map of indexed properties to be injected
	 */		
	void setIndexedProps( Map indexedProps ){
		this.indexedProps = indexedProps
	}
	
	/**
	 * Setter method for member variable testClasses
	 * @param testClasses	String list of comma separated classes to run single tests on
	 */		
	void setTestClasses( String testClasses ){
		this.testClasses = testClasses
	}

	/**
	 * Setter method for member variable defaultReportdir
	 * @param defaultReportdir	String location to write junit reports to
	 */	
	void setDefaultReportdir( String defaultReportdir ){
		this.defaultReportdir = defaultReportdir
	}
	
	/**
	 * Setter method for member variable defaultTestIncludepattern
	 * @param defaultTestIncludepattern	String representation of the file include pattern for tests
	 */		
	void setDefaultTestIncludepattern( String defaultTestIncludepattern ){
		this.defaultTestIncludepattern = defaultTestIncludepattern
	}
	
	/**
	 * Setter method for member variable defaultTestExcludepattern
	 * @param defaultTestExcludepattern	String representation of the file exclude pattern for tests
	 */		
	void setDefaultTestExcludepattern( String defaultTestExcludepattern ){
		this.defaultTestExcludepattern = defaultTestExcludepattern
	}
	
	/**
	 * Setter method for member variable defaultReportPrintsummary
	 * @param defaultReportPrintsummary	String representation of if a summary should be created
	 */	
	void setDefaultReportPrintsummary( String defaultReportPrintsummary ){
		this.defaultReportPrintsummary = defaultReportPrintsummary
	}
	
	/**
	 * Setter method for member variable defaultReportFormatter
	 * @param defaultReportFormatter	String representation of the type of reports that junit should generate
	 */	
	void setDefaultReportFormatter( String defaultReportFormatter ){
		this.defaultReportFormatter = defaultReportFormatter
	}

	/**
	 * Setter method for member variable fork
	 * @param fork	
	 */	
	void setDefaultFork( String defaultFork ){
		this.defaultFork = defaultFork
	}
	/**
	 * Setter method for member variable defaultForkMode
	 * @param defaultForkMode
	 */	
	void setDefaultForkMode( String defaultForkMode ){
		this.defaultForkMode = defaultForkMode
	}
	
	/**
	 * Sets the pre classpath
	 * @param preClasspathId
	 */
	void setPreClasspathId( String preClasspathId ) {
		this.preClasspathId = preClasspathId
	}

	/**
	 * Sets the post classpath for the JUnit id
	 * @param postClasspathId
	 */
	void setPostClasspathId( String postClasspathId ) {
		this.postClasspathId = postClasspathId
	}	
	
	/**
	 * Returns the system key options that can be set. This is similar to indexed environment variables.
	 * The entries should be in the format of the following cobertura integration example:
	 *   junit.syskey.0.key=net.sourceforge.cobertura.datafile
	 *   junit.syskey.0.file=basedir/cobertura/mydatafile
	 *   
	 *   key is mandatory but the other part of the pair may be value or path
	 */
	protected List getSysKeyOptions() {
		List sysKeyOptions = (indexedProps.containsKey("syskey")) ? indexedProps.get("syskey") : new ArrayList()
	}
	
	/**
	 * Batch test functionaility.  Loops through project test properties and creates batch tests for each using 
	 * any speciefed test properties or if none exist using the plugin defaults. 
	 * @param testProperty	 		List of Project test properties
	 */
	def batchTests( testProperty, junitTestProperty ) {

		// Define the single instance for this batch test
		Expando junitProperty
		
		// Fetch the instance from the set of indexed properties.
		if ( junitTestProperty != null )
			junitProperty = junitTestProperty
		else 
			// Else there is an issue as there are no properties to work with... 
			junitProperty = new Expando()				
						
		// check properties are present and if not set to default				
		if ( junitProperty.reportdir == null ){ 
			junitProperty.reportdir = defaultReportdir
		}
		if ( junitProperty.includepattern == null ){ 
			junitProperty.includepattern = defaultTestIncludepattern
		}
		if ( junitProperty.excludepattern == null ){ 
			junitProperty.excludepattern = defaultTestExcludepattern
		}
		if ( junitProperty.printsummary == null ){ 
			junitProperty.printsummary =  defaultReportPrintsummary
		}
		if ( junitProperty.reportformatter == null ){ 
			junitProperty.reportformatter = defaultReportFormatter
		}
		if ( junitProperty.fork == null ){ 
			junitProperty.fork = defaultFork
		}
		if ( junitProperty.forkMode == null ){ 
			junitProperty.forkMode = defaultForkMode
		}		
		// Output dir taken from the project properties
		junitProperty.dir = testProperty.output
		
		// Output the details of the batch test if debug is on.
		log.debug( batchTestPrint( junitProperty ) )
		
		// Make the reporting dir, just in case it doesn't exist.
		ant.mkdir( dir:junitProperty.reportdir )

		// Set the Junit Task options
		def junitOptions = [printsummary:junitProperty.printsummary, fork:junitProperty.fork, forkmode:junitProperty.forkMode]

		// Run Junit..
		ant.junit( junitOptions ) {
			
			// Set any env system options.
			if(getSysKeyOptions() != null && getSysKeyOptions().size() > 0) {
				getSysKeyOptions().each() { option ->
					// Check for each of the three options that can exist in a sys key
					// and write out the apporiate ant
					if( option.getProperties().containsKey("file") ) {
						ant.sysproperty(key: option.key, file: option.file)
					}
					else if( option.getProperties().containsKey("value") ) {
						ant.sysproperty(key: option.key, value: option.value)
					}
					else if( option.getProperties().containsKey("path") ) {
						ant.sysproperty(key: option.key, path: option.path)
					} 
				}
			}

			// Set the junit pre classpath
			if( this.preClasspathId != null && !this.preClasspathId.equals( "" ) ) {
				ant.classpath( refid: this.preClasspathId )
			}
			
			// Set the junit path
			ant.classpath( refid:"junit.classpath" ) 

			// Set the junit post classpath
			if( this.postClasspathId != null && !this.postClasspathId.equals( "" ) ) {
				ant.classpath( refid: this.postClasspathId )
			}

			// Set the formats
			ant.formatter( type:junitProperty.reportformatter )
			
			// Set the batch test to run.. 
			ant.batchtest( todir:junitProperty.reportdir ){										
				ant.fileset( dir:junitProperty.dir ) {
				           ant.include( name:junitProperty.includepattern )
				           ant.exclude( name:junitProperty.excludepattern )					   
				}
			}					
		}
	}
	
	/**
	 * Create formatted string of batch test variables for debugging
	 * @param junitProperty - Individual Batch test setup variables
	 */
	def String batchTestPrint( Expando junitProperty )
	{
		String outString =   "\n******  Batch Test  ******* \n"
		outString +=  "source out: $junitProperty.dir \n" ; 
		outString +=  "report out: $junitProperty.reportdir \n" ; 
		outString +=  "include: $junitProperty.includepattern  \n" ; 
		outString +=  "exclude: $junitProperty.excludepattern \n" ; 
		outString +=  "print summary: $junitProperty.printsummary \n" ; 
		outString +=  "report format: $junitProperty.reportformatter \n" ; 
		outString +=  "***************************\n" ; 	
		return outString
	}
	
	/** 
	 * Defines class path variables for use during testing.
	 * @param projectProperties	 Map of project properties
	 */
	def defineClasspath( projectProperties )
	{
		def artifactId = buildContext.get( "artifactid" )
		def srcDirs = projectProperties.get( "src" )		
		
		//groovy.util.AntBuilder monkey = new AntBuilder(ant.project)
		if ( log.isDebugEnabled() ) log.debug( "define junit.classpath" )
		ant.path( id:"junit.classpath" ) {
			
			srcDirs.each{ srcDir ->
				ant.pathelement( location: srcDir.output )
			}
			
			if( buildContext.isSet( "${artifactId}.compile.classpath" ) ){
				ant.path( refid:"global.${artifactId}.compile.classpath" )
			}
			if( buildContext.isSet( "${artifactId}.test.classpath" ) ){
				ant.path( refid:"global.${artifactId}.test.classpath" )
			}
			if( buildContext.isSet( "${artifactId}.extra.classpath" ) ){
				ant.path( refid:"global.${artifactId}.extra.classpath" )
			}
			
			log.debug "Processing project test properties of " + projectProperties["test"] 
			
			// Add all of the project test output paths to the classpath
			projectProperties["test"].each{ propertyObject ->
				// Check to see if there is a specified output directory
				ant.pathelement( path : "$propertyObject.output" )
			}
		}
		 
		  if ( log.isDebugEnabled() ) 
		  {
		  	def jclasspath = ant.project.getReference( "junit.classpath" )
		  	log.debug( "Classpath is $jclasspath" )
		  }
		
	}
	
	/*
	 * Concrete implementation of the AbstractDelegate.execute() method
	 */
	void doExecute() {
		
		log.debug( "Starting junit plugin" )

		// Fetch the global project properties
		Map projectProperties = buildContext.get( "project.properties" )
		// Now fetch the properties for all of the test source directories..
		List projectTestProperties = projectProperties.get( "test" ) 
		
		// Set the required classpath
		defineClasspath( projectProperties )
		
		// If there are test classes declared for one off test runs then run them first.  
		if( testClasses != null && testClasses != "" ) {
			// Split the comma seperated list of tests up
			List tests = testClasses.split( ',' )
			
			log.info( "$tests.size individual tests have been identified. Preparing to execute.")
			log.debug( "Individual tests identified as $tests" )
			
			// test each class defined
			tests.each { value ->	
			
				// Set the Junit Task options
				def junitOptions = [printsummary:this.defaultReportPrintsummary, fork:this.defaultFork, forkmode:this.defaultForkMode]

				log.info( "Running test: $value with options $junitOptions" )
				
				ant.junit( junitOptions ){

					// Set any env system options.
					if(getSysKeyOptions() != null && getSysKeyOptions().size() > 0) {
						getSysKeyOptions().each() { option ->
							// Check for each of the three options that can exist in a sys key
							// and write out the apporiate ant
							if( option.getProperties().containsKey("file") ) {
								ant.sysproperty(key: option.key, file: option.file)
							}
							else if( option.getProperties().containsKey("value") ) {
								ant.sysproperty(key: option.key, value: option.value)
							}
							else if( option.getProperties().containsKey("path") ) {
								ant.sysproperty(key: option.key, path: option.path)
							} 
						}
					}

					// Set the junit pre classpath
					if( this.preClasspathId != null && !this.preClasspathId.equals( "" ) ) {
						ant.classpath( refid: this.preClasspathId )
					}
					
					// Set the junit path
					ant.classpath( refid:"junit.classpath" ) 

					// Set the junit post classpath
					if( this.postClasspathId != null && !this.postClasspathId.equals( "" ) ) {
						ant.classpath( refid: this.postClasspathId )
					}
					
					ant.formatter( type:defaultReportFormatter )
					ant.test( name:value )
				}
			}
		}

		// If there are any batch test data declared then run now run them.. 
		if( projectTestProperties != null &&  this.indexedProps != null && indexedProps.containsKey( "test" ) ) {
			
			// Fetch the list of test properites.
			List junitTestProperties = indexedProps.get("test")
			
			log.info( "$projectTestProperties.size batch tests have been identified. Preparing to execute.")
			
			// For each set of indexed properties, 
			projectTestProperties.eachWithIndex() { projectTestProperty, index ->
				// Run the batch test with this property
				// The combination of the project metadata and the matching junit metadata for that index will be used.
				batchTests( projectTestProperty, junitTestProperties[index] )
			}
		}
	}
}